/*
  Copyright (C) 2008 Sebastian Reitenbach

  This file is part of OpenGroupware.org.

  OGo is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  OGo is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with OGo; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/

#include <OGoFoundation/OGoComponent.h>
#include <stdlib.h>

@class EODataSource;

@interface OGoAddressMap : OGoComponent
{
  EODataSource *dataSource;
  NSString *tempFileName;
  id labels;
}
- (void)setDataSource:(EODataSource *)_dataSource;
- (EODataSource *)dataSource;

- (NSString *)googleMapsURL;
- (NSString *)xmlFileWebURL;
- (NSString *)centerCoordinates;

- (NSString *)_xmlTempFileName;
- (NSString *)_xmlFQTempFileName;
- (NSString *)_createMarkerForCompany:(id)_company;
- (BOOL)_createXMLFileAtPath:(NSString *)_fileName;
- (NSString *) _createXMLContents;


@end

#include "common.h"
#include <OGoContacts/SkyCompanyDocument.h>
#include <OGoContacts/SkyAddressDocument.h>
#include <LSFoundation/NSObject+Commands.h>

@implementation OGoAddressMap

static NSString     *GoogleMapsURL;
static NSString     *GoogleMapsXMLCoordinatesFileBaseURL;
static NSString     *GoogleMapsXMLCoordinatesWebBaseURL;
static NSString     *GoogleMapsCenterCoordinates;

+ (NSInteger)version {
  return [super version] + 0 /* v4 */;
}
+ (void)initialize {
  NSAssert2([super version] == 2,
            @"invalid superclass (%@) version %i !",
            NSStringFromClass([self superclass]), [super version]);

  NSUserDefaults *ud = [NSUserDefaults standardUserDefaults];

  GoogleMapsURL = [[ud objectForKey:@"GoogleMapsURL"] copy];
  GoogleMapsXMLCoordinatesFileBaseURL = [[ud objectForKey:@"GoogleMapsXMLCoordinatesFileBaseURL"] copy];
  GoogleMapsXMLCoordinatesWebBaseURL = [[ud objectForKey:@"GoogleMapsXMLCoordinatesWebBaseURL"] copy];
  GoogleMapsCenterCoordinates = [[ud objectForKey:@"GoogleMapsCenterCoordinates"] copy];

}

- (id)init {
  if ((self = [super init])) {
    self->tempFileName = [self _xmlTempFileName];
  }
  return self;
}


- (void)dealloc {
  [self->labels 	release];
  [self->dataSource    	release];
  [self->tempFileName  	release];
  [super dealloc];
}

/* notifications */

- (void)sleep {
  [self->labels release]; self->labels = nil;
  [super sleep];
}

- (NSString *)_xmlTempFileName {
  int randomNumber;
  NSString *tFileName;

  srand((unsigned int) [NSDate dateWithTimeIntervalSince1970:0]);
  randomNumber = rand();
  tFileName = [[NSMutableString alloc] initWithFormat:@"coordinates-%d.xml",randomNumber];
  
  return tFileName;
}

- (NSString *)_xmlFQTempFileName {
  NSString *fqFileName;

  fqFileName = [[NSString alloc] initWithFormat:@"%@/%@", 
		GoogleMapsXMLCoordinatesFileBaseURL,
		self->tempFileName];
  return fqFileName;
}

- (NSString *)_createMarkerForCompany:(id)_company {
  LSCommandContext *cmdCtx;
  NSArray *addresses;
  NSEnumerator *e;
  NSString *fullName;
  NSString *email;

  if ([_company respondsToSelector:@selector(firstname)]) {
    fullName = [[NSString alloc]initWithFormat:@"%@ %@", [_company valueForKey:@"firstname"], [_company name]];
    email = [[NSString alloc] initWithString:[_company valueForKey:@"email1"]?[_company valueForKey:@"email1"]:@""];
  } else {
    fullName = [[NSString alloc]initWithString:[_company name]];
    email = [[NSString alloc] initWithString:[_company valueForKey:@"email"]?[_company valueForKey:@"email"]:@""];
  }

  
  id address;
  NSMutableString *marker;

  marker = [[NSMutableString alloc] init];
  cmdCtx = [(NSObject *)[self session] commandContext];
  [marker appendString:@""];
  if (![cmdCtx begin]) {
    [self errorWithFormat:@"could not begin cmdctx tx!"];
  } else {
    addresses = [cmdCtx runCommand:@"address::get",
		@"companyId", [_company valueForKey:@"companyId"],
		@"operator", @"AND", 
		@"returnType", intObj(LSDBReturnType_ManyObjects), nil];

    e = [addresses objectEnumerator];

    while ((address = [e nextObject])) {
      NSDictionary *coordsDict;
      coordsDict = (NSDictionary *)[cmdCtx runCommand:@"address::get-geocode",
                 @"address", address, nil];
      if ([[coordsDict valueForKey:@"latitude"] length] > 0 && 
	[[coordsDict valueForKey:@"longitude"] length] > 0) {
        [marker appendFormat:@"<marker lat=\"%@\" lng=\"%@\" html=\"&lt;b&gt;%@&lt;/b&gt;&lt;br&gt;\
			Email: &lt;a href=&quot;mailto:%@&quot;&gt;%@&lt;/a&gt;&lt;br&gt;\
			URL: &lt;a href=&quot;%@&quot;&gt;%@&lt;/a&gt;\
			&lt;table&gt;&lt;tr&gt;&lt;th&gt;Address:&lt;/th&gt;&lt;td&gt;%@&lt;/td&gt;&lt;/tr&gt;\
			&lt;tr&gt;&lt;th&gt;Street:&lt;/th&gt;&lt;td&gt;%@&lt;/td&gt;&lt;/tr&gt;\
			&lt;tr&gt;&lt;th&gt;City:&lt;/th&gt;&lt;td&gt;%@&lt;/td&gt;&lt;/tr&gt;\
			&lt;tr&gt;&lt;th&gt;ZIP:&lt;/th&gt;&lt;td&gt;%@&lt;/td&gt;&lt;/tr&gt;\
			&lt;tr&gt;&lt;th&gt;State:&lt;/th&gt;&lt;td&gt;%@&lt;/td&gt;&lt;/tr&gt;\
			&lt;tr&gt;&lt;th&gt;Country:&lt;/th&gt;&lt;td&gt;%@&lt;/td&gt;&lt;/tr&gt;\
			&lt;tr&gt;&lt;th&gt;Telephone:&lt;/th&gt;&lt;td&gt;%@&lt;/td&gt;&lt;/tr&gt;\
			&lt;tr&gt;&lt;th&gt;Mobile:&lt;/th&gt;&lt;td&gt;%@&lt;/td&gt;&lt;/tr&gt;\
			&lt;/table&gt;\
			\" label=\"%@\" />",
		[coordsDict valueForKey:@"latitude"], 
		[coordsDict valueForKey:@"longitude"],
		fullName, 
		email,
		email,
		[_company valueForKey:@"url"],
		[_company valueForKey:@"url"],
		[address valueForKey:@"type"],
		[address valueForKey:@"street"],
		[address valueForKey:@"city"],
		[address valueForKey:@"zip"],
		[address valueForKey:@"state"],
		[address valueForKey:@"country"],
		[_company valueForKey:@"01_tel"],
		[_company valueForKey:@"03_tel_funk"],
		fullName];
	break;
      }
    } 
  }
  // commit the open transaction
  if ([cmdCtx isTransactionInProgress]) {
    if (![cmdCtx commit])
      NSLog(@"OGoAddressMap.m: could not commit transaction!");
  }
  return marker;
}

- (NSString *)_createXMLContents {
  id companies;
  id company;
  NSEnumerator *e;
  NSMutableString *contents;

  companies = [[self dataSource] fetchObjects];
  contents = [[NSMutableString alloc] init];
  [contents appendString:@"<markers>\n"];
  e = [companies objectEnumerator];
 
  while ((company = [e nextObject])) {
    [contents appendFormat:@"  %@\n", [self _createMarkerForCompany:company]];
  }
  [contents appendString:@"</markers>\n"];
  
  return contents;
}

- (BOOL) _createXMLFileAtPath:(NSString *)_fileName {
  NSString *contents;
  NSFileManager *defaultManager;
  defaultManager = [NSFileManager defaultManager];
  
  [defaultManager removeFileAtPath:_fileName handler: nil];
  contents = [[NSString alloc] initWithString:[self _createXMLContents]];
  if ([contents writeToFile:_fileName atomically:YES])
    return YES;
  else
    return NO;
}

/* accessors */

- (void)setDataSource:(EODataSource *)_dataSource {
  ASSIGN(self->dataSource, _dataSource);
}
- (EODataSource *)dataSource {
  return self->dataSource;
}


- (void)setLabels:(id)_labels {
  ASSIGN(self->labels, _labels);
}
- (id)labels {
  return self->labels;
}

- (NSString *)centerCoordinates {
  return GoogleMapsCenterCoordinates;
}

- (NSString *)urlForResourceNamed:(NSString *)_name {
  WOResourceManager *rm;
  NSString *url;
  NSArray  *langs;

  langs = [self hasSession]
    ? [[self session] languages]
    : [[[self context] request] browserLanguages];

  rm = [self resourceManager];
  url = [rm urlForResourceNamed:_name inFramework:nil
            languages:langs request:[[self context] request]];
  return url;
}
- (NSString *)stylesheetURL {
  return [self urlForResourceNamed:@"OGo.css"];
}

- (NSString *)xmlFileWebURL {
  NSString *webURL;

  // first create the xml file before creating the URL;
  if ([self _createXMLFileAtPath:[self _xmlFQTempFileName]]) {
    webURL = [[NSString alloc] initWithFormat:@"%@/%@", 
		GoogleMapsXMLCoordinatesWebBaseURL,
		self->tempFileName];
  } else {
    webURL = [[NSString alloc] initWithString:@""];
  }
  return webURL;
}

- (NSString *)googleMapsURL {
  return GoogleMapsURL;
}

@end /* OGoAddressMap */
